from datetime import datetime
from pathlib import Path
from typing import Any, Dict, Mapping, Optional, Sequence, Tuple, Union

from opa_tools.config import REQUEST_DIR


ALLOWED_KINDS = (
    "gsv",
    "opa",
    "par",
)

RequestType = Mapping[str, Any]

class GenericRequestError(ValueError):
    def __init__(self, *args):
        self.identifier = args[0]
        msg = (
            f"Request {self.identifier} contains subrequests. "
            "Please, specify request in one of the following forms: "
            "{'MM/REQID.SUBREQID', 'YYMM/REQID.SUBREQID', 'REQID.SUBREQID'}."
        )
        super().__init__(msg, *args[1:])

class RequestIdentifierError(ValueError):
    def __init__(self, *args):
        self.identifier = args[0]
        msg = (
            f"Malformed request: {self.identifier}. "
            "It should be in one of the following forms: "
            "{'MM/REQID', 'YYMM/REQID', 'REQID'}."
        )
        super().__init__(msg, *args[1:])

# TODO: Implement msg for either 1 or 4 args
class RequestNotFoundError(ValueError):
    def __init__(self, *args):
        msg = (
            f"Path for request not found: {args[0]}."
        )
        super().__init__(msg, *args[1:])

def _normalize_request_name(name: str):
    return f"{int(name):03d}"

def _normalize_subrequest_name(subrequest: Union[str, None]):
    if subrequest is None:
        return None
    return str(subrequest).lstrip("0")

def _normalize_request_name_subrequest(name: Any, subrequest: Optional[Any] = None) -> Tuple[str, Union[str, None]]:
    if isinstance(name, float):
        raise NotImplementedError(
            "Passing request name as float not implemented yet. "
            "Please, pass as a string with a dot."
        )
    if subrequest is not None:
        subrequest = str(subrequest)
    if isinstance(name, int):
        name = str(name)
    elif isinstance(name, str):
        if "." in name:
            name, subrequest = name.split(".", 1)
    else:
        raise TypeError(
            "Request/subrequest names must be expressed as "
            f"int, float or str, but were {type(name)}/{type(subrequest)}."
        )
    return _normalize_request_name(name), _normalize_subrequest_name(subrequest)

def _validate_month(month: str):
    if not 1 <= int(month) <= 12:
        raise ValueError(f"Bad value for month: {month}.")

def _normalize_request_month(month: Union[str, int, None]) -> str:
    now = datetime.now()
    if month is None:
        return now.strftime("%y%m")
    month = str(month)
    if month == "*":
        return month
    if not month.isdigit():
        raise ValueError(f"Value for month must be a number, but was '{month}'.")
    if len(month) == 2:
        _validate_month(month)
        return now.strftime("%y") + month
    elif len(month) == 4:
        _validate_month(month[2:4])
        return month
    raise ValueError(
        f"Month must be of length either 2 or 4, was '{month}'."
    )

def normalize_request_identifier(
    identifier: Any,
    /,
    subrequest: Optional[Any] = None,
    month: Optional[Union[str, int]] = None,
) -> Tuple[str, Union[str, None], str]:
    # Split identifier
    if isinstance(identifier, str) and "/" in identifier:
        month, name = identifier.split("/", 1)
    else:
        name = identifier
        try:
            month = _normalize_request_month(month)
        except ValueError as exc:
            raise RequestIdentifierError(identifier) from exc
    name, subrequest = _normalize_request_name_subrequest(name, subrequest)
    try:
        month = _normalize_request_month(month)
    except ValueError as exc:
        raise RequestIdentifierError(identifier) from exc
    return name, subrequest, month

def format_identifier(*args):
    if len(args) == 1:
        yymm = datetime.now().strftime("%y%m")
        return "{}/{}".format(yymm, *args)
    args = (_normalize_request_month(args[0]), *args[1:])
    if len(args) == 2:
        return "{}/{}".format(*args)
    if len(args) == 3:
        return "{}/{}.{}".format(*args)
    raise ValueError(
        f"1 to 3 args expected, {len(args)} passed."
    )

def validate_request_kind(kind: Union[str, None]):
    if kind is None:
        return
    if kind not in ["*", *ALLOWED_KINDS]:
        raise ValueError(
            f"Wrong kind: '{kind}', must be in {ALLOWED_KINDS}."
        )


def _find_zero_padded_subdir(path: Path, subdir: str, max_digits=12):
    if len(subdir) >= max_digits:
        raise RuntimeError(
            f"Subdirectory {subdir} too long to be found in {path}"
        )
    for n_zeros in range(0, max_digits - len(subdir)):
        cur_path = path / subdir.zfill(n_zeros)
        if cur_path.exists():
            return cur_path
    raise FileNotFoundError(
        f"Directory {path / subdir} (or any other zero-padded variant, "
        f"up to {max_digits} digits) not found for request."
    )

def _get_subrequests(
        request_path: Path,
        no_exist_ok: bool = False
    ) -> Dict:
    raise NotImplementedError(
        "Ability to fetch lists of subrequests not implemented yet. "
        "Please, pass subrequests manually (syntax: YYMM/REQID.SUBREQID)"
    )


def get_request_location(
    identifier: Optional[Union[str, int, float]] = None,
    /,
    subrequest: Optional[str] = None,
    month: Optional[str] = None,
    kind: Optional[str] = None,
    *,
    allow_subrequests: bool = False,
    no_exist_ok: bool = False,
) -> Path:
    if kind not in [None, *ALLOWED_KINDS]:
        raise ValueError(
            f"kind must be either in {ALLOWED_KINDS}, but was '{kind}'."
        )
    name, subrequest, month = normalize_request_identifier(
        identifier, subrequest, month
    )
    request_dir = _find_zero_padded_subdir(REQUEST_DIR / month, name)
    if not no_exist_ok and not request_dir.exists():
        raise RequestNotFoundError(request_dir)
    if subrequest:
        request_dir = _find_zero_padded_subdir(request_dir, subrequest)
        if not no_exist_ok and not request_dir.exists():
            raise RequestNotFoundError(request_dir)
    if kind is None:
        return request_dir
    request_path = request_dir / f"{kind}.yml"
    if request_path.exists():
        return request_path
    # From here on, request path does not exist
    if any(subpath.is_dir() for subpath in request_dir.iterdir()):
        if allow_subrequests:
            return _get_subrequests(request_path, no_exist_ok=no_exist_ok)
        raise GenericRequestError(name, identifier)
    if no_exist_ok:
        return request_path
    raise RequestNotFoundError(request_path)
    
